#include "cron.h"
#include <selinux/context.h>
#include <selinux/get_context_list.h>
#include <selinux/flask.h>
#include <selinux/av_permissions.h>

static int
selinux_authorize(security_context_t scon, security_context_t tcon,
		security_class_t tclass,  access_vector_t requested) {
	struct av_decision avd;
	int rc;

	rc = security_compute_av(scon, tcon, tclass, requested, &avd);
	if (rc || ((requested & avd.allowed) != requested))
		return 0;
	return 1;
}

static int
selinux_authorize_context(security_context_t scontext,
		security_context_t file_context) {
	/* Since crontab files are not directly executed,
	 * crond must ensure that the crontab file has
	 * a context that is appropriate for the context of
	 * the user cron job.  It performs an entrypoint
	 * permission check for this purpose.
	 */
	return selinux_authorize(scontext, file_context, SECCLASS_FILE, FILE__ENTRYPOINT);
}

static int
selinux_authorize_range(const security_context_t scontext,
		const security_context_t ucontext) {
	/* Since crontab files are not directly executed,
	 * so crond must ensure that any user specified range
	 * falls within the seusers-specified range for that Linux user.
	 */
	return selinux_authorize(scontext, ucontext, SECCLASS_CONTEXT, CONTEXT__CONTAINS);
}

static security_context_t
get_job_context(const char *u_name, security_context_t u_scontext, const char *range)
{
	security_context_t scontext = NULL;
	context_t ccon = NULL;

	if (range) {
		if (!(ccon = context_new(u_scontext))) {
			log_it(u_name, getpid(),
				"context_new FAILED for MLS_LEVEL", range);
			return NULL;
		}

		if (context_range_set(ccon, range)) {
			log_it(u_name, getpid(),
				"context_range_set FAILED for MLS_LEVEL", range);
			goto out;
		}

		if (!(scontext = context_str(ccon))) {
			log_it(u_name, getpid(),
				"context_str FAILED for MLS_LEVEL", range);
			goto out;
		}
	}

	if (!(scontext = strdup(scontext ? scontext : u_scontext))) {
		log_it(u_name, getpid(), "strdup FAILED for MLS_LEVEL", range);
	}

out:
	context_free(ccon);
	return scontext;
}

security_context_t
get_selinux_context(const char *name, int fd, char **err_msg) {
	security_context_t user_context = NULL;
	security_context_t file_context = NULL;
	char *seuser = NULL;
	char *level = NULL;

	if (fgetfilecon(fd, &file_context) < OK) {
		*err_msg = strdup("getfilecon FAILED");
		return NULL;
	}

	if (name && getseuserbyname(name, &seuser, &level) < OK) {
		*err_msg = strdup("NO SEUSER");
		goto out;
	}

	if (get_default_context_with_level(name ? seuser : "system_u",
					level, NULL, &user_context) < OK) {
		*err_msg = strdup("No SELinux security context");
		goto out;
	}

	if (!selinux_authorize_context(user_context, file_context)) {
		freecon(user_context);
		user_context = NULL;
		*err_msg = strdup("Unauthorized SELinux context");
	}

out:
	free(seuser);
	free(level);
	freecon(file_context);

	return user_context;
}

int
set_job_selinux_context(const user *u, char **envp)
{
	security_context_t scontext = NULL;
	int rc = -1;

	if (u->scontext == NULL) {
		log_it(u->name, getpid(), "NULL security context for user", "");
		return -1;
	}

	scontext = get_job_context(u->name, u->scontext, env_get("MLS_LEVEL", envp));
	if (!scontext)
		return -1;
	if (strcmp(u->scontext, scontext) &&
			!selinux_authorize_range(u->scontext, scontext)) {
		char *msg = NULL;
		if (asprintf(&msg, "Unauthorized range in %s for user range in %s",
				(char *) scontext, (const char *)u->scontext) >= 0) {
			log_it(u->name, getpid(), "ERROR", msg);
			free(msg);
		}
		goto out;
	}

	if (setexeccon(scontext) < 0 || setkeycreatecon(scontext) < 0) {
		char *msg = NULL;
		if (asprintf(&msg, "Could not set exec or keycreate context to %s for user",
				(char *) scontext) >= 0) {
			log_it(u->name, getpid(), "ERROR", msg);
			free(msg);
		}
		goto out;
	}

	rc = 0;
out:
	freecon(scontext);

	return rc;
}

int
set_at_selinux_context(const char *name, int fd, char **err_msg)
{
	security_context_t scontext = NULL;
	int rc = 0;

	scontext = get_selinux_context(name, fd, err_msg);
	if (scontext == NULL)
		return -1;

	if (setexeccon(scontext) < 0 || setkeycreatecon(scontext) < 0) {
		asprintf(err_msg, "Could not set exec or keycreate context to %s",
					(char *) scontext);
		rc = -1;
	}

	freecon(scontext);

	return rc;
}

int
selinux_crontab_access_denied(void) {
	security_class_t passwd_class;
	security_context_t scontext;
	int rc = 1;

	if (is_selinux_enabled() <= 0)
		return 0;

	if (getprevcon(&scontext)) {
		perror("Cannot obtain SELinux process context");
		return security_getenforce() > 0;
	}

	passwd_class = string_to_security_class("passwd");
	if (passwd_class <= 0) {
		fprintf(stderr, "Security class \"passwd\" is not defined in the SELinux policy.\n");
		goto out;
	}

	rc = !selinux_authorize(scontext, scontext, passwd_class, PASSWD__CRONTAB);

out:
	freecon(scontext);

	return rc ? (security_getenforce() > 0) : 0;
}

int
write_crontab_mls_level(FILE *crontab) {
	security_context_t scontext = NULL;
	context_t ccon = NULL;
	const char *level = NULL;
	int rc = -1;

	if (getprevcon(&scontext)) {
		perror("Cannot obtain SELinux process context");
		return -1;
	}

	if (!(ccon = context_new(scontext))) {
		perror("context_new failed");
		goto out;
	}

	if (!(level = context_range_get(ccon))) {
		perror("context_range failed");
		goto out;
	}

	rc = fprintf(crontab, "MLS_LEVEL=%s\n", level);
	if (rc < 0)
		perror("Cannot write MLS level");

out:
	context_free(ccon);
	freecon(scontext);

	return rc;
}

